"""
Load local .env from CWD or path, if provided

Current format for the `.env` file supports strings only and is parsed in
the following order:
- Each seperate line is considered a new possible key/value set
- Each set is delimted by the first `=` found
- Leading and trailing whitespace are removed
- Matched leading/trailing single quotes or double quotes will be
  stripped from values (not keys).

I'm open to suggestions on standards to follow here.

Author  : Preocts <Preocts#8196>
Git Repo: https://github.com/Preocts/secretbox
"""
import logging
import re

from secretbox.loader import Loader


class EnvFileLoader(Loader):
    """Load local .env file"""

    LT_DBL_QUOTES = r'^".*"$'
    LT_SGL_QUOTES = r"^'.*'$"
    EXPORT_PREFIX = r"^\s*?export\s"

    logger = logging.getLogger(__name__)

    def load_values(self, **kwargs: str) -> bool:
        """
        Loads local .env from cwd or path, if provided

        Keywords:
            filename : [str] Alternate filename to load over `.env`
        """
        filename = kwargs.get("filename", ".env")
        self.logger.debug("Reading vars from '%s'", filename)
        try:
            with open(filename, "r", encoding="utf-8") as input_file:
                self.parse_env_file(input_file.read())
        except FileNotFoundError:
            return False
        return True

    def parse_env_file(self, input_file: str) -> None:
        """Parses env file into key-pair values"""
        for line in input_file.split("\n"):
            if not line or line.strip().startswith("#") or len(line.split("=", 1)) != 2:
                continue
            key, value = line.split("=", 1)

            key = self.strip_export(key).strip()
            value = value.strip()

            if value.startswith('"'):
                value = self.remove_lt_dbl_quotes(value)
            elif value.startswith("'"):
                value = self.remove_lt_sgl_quotes(value)

            self.loaded_values[key] = value

    def remove_lt_dbl_quotes(self, in_: str) -> str:
        """Removes matched leading and trailing double quotes"""
        return in_.strip('"') if re.match(self.LT_DBL_QUOTES, in_) else in_

    def remove_lt_sgl_quotes(self, in_: str) -> str:
        """Removes matched leading and trailing double quotes"""
        return in_.strip("'") if re.match(self.LT_SGL_QUOTES, in_) else in_

    def strip_export(self, in_: str) -> str:
        """Removes leading 'export ' prefix, case agnostic"""
        return re.sub(self.EXPORT_PREFIX, "", in_, flags=re.IGNORECASE)



from abc import ABC
from typing import Any
from typing import Dict


class Loader(ABC):
    """Abstract Base Class for all loaders"""

    def __init__(self) -> None:
        super().__init__()
        self.loaded_values: Dict[str, str] = {}

    def load_values(self, **kwargs: Any) -> bool:
        """Override with loading optionation, store within self.loaded_values"""
        raise NotImplementedError()  # pragma: no cover



"""
Load system environ values

Author  : Preocts <Preocts#8196>
Git Repo: https://github.com/Preocts/secretbox
"""
import logging
import os

from secretbox.loader import Loader


class EnvironLoader(Loader):
    """Load environ values"""

    logger = logging.getLogger(__name__)

    def load_values(self, **kwargs: str) -> bool:
        """Loads all visible environmental variables"""
        self.logger.debug("Reading %s environ variables", len(os.environ))
        self.loaded_values.update(os.environ)
        return True



"""
Load secrets from an AWS secret manager

Author  : Preocts <Preocts#8196>
Git Repo: https://github.com/Preocts/secretbox
"""
import json
import logging
import os
from typing import Any
from typing import Dict
from typing import Optional

try:
    import boto3
    from botocore.exceptions import ClientError
    from botocore.exceptions import InvalidRegionError
    from botocore.exceptions import NoCredentialsError
    from botocore.exceptions import NoRegionError
    from mypy_boto3_secretsmanager.client import SecretsManagerClient
except ImportError:
    boto3 = None
    SecretsManagerClient = None

from secretbox.loader import Loader


class AWSSecretLoader(Loader):
    """Load secrets from an AWS secret manager"""

    logger = logging.getLogger(__name__)

    def __init__(self) -> None:
        super().__init__()
        self.aws_sstore: Optional[str] = None
        self.aws_region: Optional[str] = None

    def populate_region_store_names(self, **kwargs: Any) -> None:
        """Populates store/region name"""
        kw_sstore = kwargs.get("aws_sstore_name")
        kw_region = kwargs.get("aws_region_name")
        os_sstore = os.getenv("AWS_SSTORE_NAME")
        os_region = os.getenv("AWS_REGION_NAME")

        # Use the keyword over the os, default to None
        self.aws_sstore = kw_sstore if kw_sstore is not None else os_sstore
        self.aws_region = kw_region if kw_region is not None else os_region

    def load_values(self, **kwargs: Any) -> bool:
        """
        Load all secrets from AWS secret store
        Requires `aws_sstore_name` and `aws_region_name` keywords to be
        provided or for those values to be in the environment variables
        under `AWS_SSTORE_NAME` and `AWS_REGION_NAME`.

        `aws_sstore_name` is the name of the store, not the arn.
        """
        if boto3 is None:
            self.logger.debug("Skipping AWS loader, boto3 is not available.")
            return False

        self.populate_region_store_names(**kwargs)
        secrets: Dict[str, str] = {}
        aws_client = self.connect_aws_client()

        if aws_client is None or self.aws_sstore is None:
            self.logger.debug("Cannot load AWS secrets, no valid client.")
            return False

        try:
            response = aws_client.get_secret_value(SecretId=self.aws_sstore)

        except NoCredentialsError as err:
            self.logger.error("Error routing message! %s", err)

        except ClientError as err:
            code = err.response["Error"]["Code"]
            self.logger.error("ClientError: %s, (%s)", err, code)

        else:
            self.logger.debug("Found %s values from AWS.", len(secrets))
            secrets = json.loads(response.get("SecretString", "{}"))
            self.loaded_values.update(secrets)
        return bool(secrets)

    def connect_aws_client(self) -> Optional[SecretsManagerClient]:
        """Make connection"""
        client: Optional[SecretsManagerClient] = None
        session = boto3.session.Session()

        if self.aws_region is None:
            self.logger.debug("No valid AWS region, cannot create client.")

        else:
            try:
                client = session.client(
                    service_name="secretsmanager",
                    region_name=self.aws_region,
                )
            except (ValueError, InvalidRegionError, NoRegionError) as err:
                self.logger.error("Error creating AWS Secrets client: %s", err)

        return client



"""
Loads various environment variables/secrets for use

Author  : Preocts <Preocts#8196>
Git Repo: https://github.com/Preocts/secretbox
"""
import logging
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type

from secretbox.awssecret_loader import AWSSecretLoader
from secretbox.envfile_loader import EnvFileLoader
from secretbox.environ_loader import EnvironLoader
from secretbox.loader import Loader

LOADERS: Dict[str, Type[Loader]] = {
    "envfile": EnvFileLoader,
    "environ": EnvironLoader,
    "awssecret": AWSSecretLoader,
}


class SecretBox:
    """Loads various environment variables/secrets for use"""

    logger = logging.getLogger(__name__)

    def __init__(
        self,
        *,
        auto_load: bool = False,
        debug_flag: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Initialize SecretBox

        Keyword Args:
            auto_load : If true, environment vars and `.env` will be loaded
            load_debug : When true, internal logger level is set to DEBUG

        Additional Keywords:
            These will be passed to all loaders when called
        """
        self.logger.setLevel(level="DEBUG" if debug_flag else "ERROR")
        self.logger.debug("Debug flag passed.")

        self.loaded_values: Dict[str, str] = {}
        self.kwarg_defaults = kwargs

        if auto_load:
            self.load_from(["environ", "envfile"])

    def get(self, key: str, default: Optional[str] = None) -> str:
        """Get a value by key, return default if not found or raise if no default"""
        if default is None:
            return self.loaded_values[key]
        else:
            return self.loaded_values.get(key, default)

    def get_int(self, key: str, default: Optional[int] = None) -> int:
        """Convert value by key to int."""
        if default is None:
            return int(self.get(key))
        else:
            value = self.get(key, "")
            return int(value) if value else default

    def get_list(
        self,
        key: str,
        delimiter: str = ",",
        default: Optional[List[str]] = None,
    ) -> List[str]:
        """Convert value by key to list seperated by delimiter."""
        if default is None:
            default = []
        if not default:
            return self.get(key).split(delimiter)
        else:
            value = self.get(key, "")
            return value.split(delimiter) if value else default

    def load_from(
        self,
        loaders: List[str],
        **kwargs: Any,
    ) -> None:
        """
        Runs load_values from each of the listed loader in the order they appear

        Loader options:
            environ:
                Loads the current environmental variables into secretbox.
            envfile:
                Loads .env file. Optional `filename` kwarg can override the default
                load of the current working directory `.env` file.
            awssecret:
                Loads secrets from an AWS secret manager. Requires `aws_sstore_name`
                and `aws_region_name` keywords to be provided or for those values
                to be in the environment variables under `AWS_SSTORE_NAME` and
                `AWS_REGION_NAME`. `aws_sstore_name` is not the arn.
        """
        for loader_name in loaders:
            self.logger.debug("Loading from interface: `%s`", loader_name)
            interface = LOADERS.get(loader_name)
            if interface is None:
                self.logger.error("Loader `%s` unknown, skipping", loader_name)
                continue
            loader = interface()
            loader.load_values(**self._join_kwarg_defaults(kwargs))
            self.logger.debug("Loaded %d values.", len(loader.loaded_values))
            self._update_loaded_values(loader.loaded_values)
        self._push_to_environment()

    def _update_loaded_values(self, new_values: Dict[str, str]) -> None:
        """Update/Create instance state of loaded values with new values"""
        self.loaded_values.update(new_values)

    def _push_to_environment(self) -> None:
        """Pushes loaded values to local environment vars, will overwrite existing"""
        for key, value in self.loaded_values.items():
            self.logger.debug("Push, %s : ***%s", key, value[-(len(value) // 4) :])
            os.environ[key] = value

    def _join_kwarg_defaults(self, new_kwargs: Dict[str, str]) -> Dict[str, str]:
        """Update default kwargs with specific while not mutating either"""
        return {**self.kwarg_defaults, **new_kwargs}
